import logging

log = logging.getLogger("curricula")
log.propagate = False
log.setLevel(logging.INFO)

formatter = logging.Formatter(fmt="%(asctime)s %(levelname)s: %(message)s", datefmt="%m/%d/%y %I:%M:%S %p")
handler = logging.StreamHandler()
handler.setFormatter(formatter)
log.addHandler(handler)



import time
import datetime

from decimal import Decimal
from pathlib import Path
from dataclasses import dataclass, asdict, field
from typing import Optional, List, Callable, TypeVar
from abc import ABC, abstractmethod
from functools import lru_cache

from .structure import *

TZ = datetime.timezone(offset=datetime.timedelta(seconds=time.timezone))


def deserialize_datetime(s: str) -> Optional[datetime.datetime]:
    """Deserialize our standard format."""

    if s is None:
        return None
    return datetime.datetime.strptime(s, "%Y-%m-%d %H:%M:%S").replace(tzinfo=TZ)


def serialize_datetime(d: datetime) -> Optional[str]:
    """Serialize back."""

    if d is None:
        return None
    return d.strftime("%Y-%m-%d %H:%M:%S")


T = TypeVar("T")
U = TypeVar("U")


def some(value: Optional[T], method: Callable[[T], U]) -> Optional[U]:
    """Probably a monad, right?"""

    if value is None:
        return value
    return method(value)


@dataclass(eq=False)
class Model(ABC):
    """Provide some default behaviors."""

    def dump(self) -> dict:
        return asdict(self)

    @classmethod
    @abstractmethod
    def load(cls, data: dict) -> "Model":
        """Load the model from serialized data."""


@dataclass(eq=False)
class Author(Model):
    """Name and email."""

    name: str
    email: str

    @classmethod
    def load(cls, data: dict) -> "Author":
        """Deserialize and check for errors."""

        return Author(name=data.pop("name"), email=data.pop("email"))


@dataclass(eq=False)
class ProblemGradingCategory(Model):
    """Data about weight, points, etc."""

    weight: Decimal
    points: Decimal

    # Whether it should be used
    enabled: bool = True

    # Intrinsic
    name: Optional[str] = None
    minutes: Optional[float] = None

    @classmethod
    def load(cls, data: dict) -> "ProblemGradingCategory":
        """Deserialize decimals."""

        return cls(
            enabled=data.get("enabled", True),
            name=data["name"],
            minutes=data.get("minutes"),
            weight=Decimal(data["weight"]),
            points=Decimal(data["points"]),)

    def dump(self) -> dict:
        """Use string format."""

        return dict(
            enabled=self.enabled,
            name=self.name,
            minutes=self.minutes,
            weight=str(self.weight),
            points=str(self.points),)


@dataclass(eq=False)
class ProblemGrading(Model):
    """Data for each grading method."""

    weight: Decimal = Decimal(0)
    points: Decimal = Decimal(0)

    enabled: bool = True

    automated: Optional[ProblemGradingCategory] = None
    review: Optional[ProblemGradingCategory] = None
    manual: Optional[ProblemGradingCategory] = None

    @property
    def is_automated(self) -> bool:
        return self.enabled and self.automated is not None and self.automated.enabled

    @property
    def is_review(self) -> bool:
        return self.enabled and self.review is not None and self.review.enabled

    @property
    def is_manual(self) -> bool:
        return self.enabled and self.manual is not None and self.manual.enabled

    @property
    @lru_cache(maxsize=1)
    def weight_total(self) -> Decimal:
        return sum((
            self.automated.weight if self.automated and self.automated.enabled else 0,
            self.review.weight if self.review and self.review.enabled else 0,
            self.manual.weight if self.manual and self.manual.enabled else 0))

    @property
    def percentage_automated(self) -> Decimal:
        return self.automated.weight / self.weight_total

    @property
    def percentage_review(self) -> Decimal:
        return self.review.weight / self.weight_total

    @property
    def percentage_manual(self) -> Decimal:
        return self.manual.weight / self.weight_total

    @classmethod
    def load(cls, data: dict) -> "ProblemGrading":
        """Deserialize each method."""

        if data["automated"] is not None and data["automated"].get("name") is None:
            data["automated"]["name"] = "Automated tests"
        if data["review"] is not None and data["review"].get("name") is None:
            data["review"]["name"] = "Code review"
        if data["manual"] is not None and data["manual"].get("name") is None:
            data["manual"]["name"] = "Manual grading"

        return cls(
            enabled=data.get("enabled", True),
            weight=Decimal(data["weight"]),
            points=Decimal(data["points"]),
            automated=some(data["automated"], ProblemGradingCategory.load),
            review=some(data["review"], ProblemGradingCategory.load),
            manual=some(data["manual"], ProblemGradingCategory.load),)

    def dump(self) -> dict:
        """Serialize with monad."""

        return dict(
            enabled=self.enabled,
            weight=str(self.weight),
            points=str(self.points),
            automated=some(self.automated, ProblemGradingCategory.dump),
            review=some(self.review, ProblemGradingCategory.dump),
            manual=some(self.manual, ProblemGradingCategory.dump),)


@dataclass(eq=False)
class Problem(Model):
    """All problem data."""

    # Overwrite
    short: str
    title: str

    # Defined in assignment reference
    grading: ProblemGrading = field(default_factory=ProblemGrading)

    # Most specific directory containing relevant files relative to assignment root
    relative_path: Path = field(default_factory=Path)

    # Intrinsic
    authors: List[Author] = field(default_factory=list)
    topics: List[str] = field(default_factory=list)
    notes: Optional[str] = None
    difficulty: Optional[str] = None

    # Backlink
    assignment: "Assignment" = None

    def weight(self) -> Decimal:
        """Percentage weight of the problem in the assignment."""

        if self.assignment.grading.weight() == 0:
            return Decimal(0)
        return self.grading.weight / self.assignment.grading.weight()

    @classmethod
    def load(cls, data: dict, assignment: "Assignment" = None) -> "Problem":
        """Load directly from a dictionary."""

        self = cls(
            assignment=assignment,
            short=data["short"],
            title=data["title"],
            relative_path=Path(data["relative_path"]),
            grading=ProblemGrading.load(data["grading"]),
            authors=list(map(Author.load, data["authors"])),
            topics=data["topics"],
            notes=data.get("notes"),
            difficulty=data.get("difficulty"))
        self.grading.problem = self
        return self

    def dump(self) -> dict:
        """Serialize 1:1."""

        return dict(
            short=self.short,
            title=self.title,
            relative_path=str(self.relative_path),
            grading=self.grading.dump(),
            authors=list(map(Author.dump, self.authors)),
            topics=self.topics,
            notes=self.notes,
            difficulty=self.difficulty,)


@dataclass(eq=False)
class AssignmentGrading(Model):
    """Weights and points."""

    points: int
    assignment: "Assignment" = field(default=None)

    @classmethod
    def load(cls, data: dict, assignment: "Assignment" = None) -> "AssignmentGrading":
        """Load from serialized."""

        return cls(points=data["points"], assignment=assignment)

    @lru_cache(maxsize=1)
    def weight(self) -> Decimal:
        """Compute cumulative weight of all problems."""

        return sum(problem.grading.weight for problem in self.assignment.problems)

    def dump(self) -> dict:
        """Avoid recursion."""

        return dict(points=self.points)


@dataclass(eq=False)
class AssignmentMeta(Model):
    """Metadata about an assignment."""

    built: datetime.datetime = field(default_factory=datetime.datetime.now)
    curricula: str = version

    @classmethod
    def load(cls, data: dict) -> "AssignmentMeta":
        """Deserialize datetime."""

        return cls(
            built=deserialize_datetime(data["built"]),
            curricula=data["curricula"],)

    def dump(self) -> dict:
        """Serialize the datetime here too."""

        return dict(
            built=serialize_datetime(self.built),
            curricula=version,)


@dataclass(eq=False)
class Assignment(Model):
    """Contains assignment metadata."""

    short: str
    title: str
    authors: List[Author]

    problems: List[Problem]
    grading: AssignmentGrading

    notes: Optional[str] = None
    meta: AssignmentMeta = AssignmentMeta()
    extra: Optional[dict] = None

    @classmethod
    def load(cls, data: dict, problems: List[Problem] = None) -> "Assignment":
        """Deserialize."""

        if problems is None:
            problems = list(map(Problem.load, data["problems"]))

        self = cls(
            short=data["short"],
            title=data["title"],
            authors=list(map(Author.load, data["authors"])),
            problems=problems,
            grading=AssignmentGrading.load(data["grading"]),
            extra=data.get("extra"),
            notes=data.get("notes"),
            meta=AssignmentMeta.load(data["meta"]) if "meta" in data else AssignmentMeta())

        for problem in problems:
            problem.assignment = self
        self.grading.assignment = self

        return self

    def dump(self) -> dict:
        """Dump the assignment to JSON."""

        return dict(
            short=self.short,
            title=self.title,
            authors=[author.dump() for author in self.authors],
            problems=[problem.dump() for problem in self.problems],
            grading=self.grading.dump(),
            extra=self.extra,
            notes=self.notes,
            meta=self.meta.dump())



from pathlib import Path
from typing import Iterator

__all__ = (
    "version",
    "Paths",
    "Files",
    "InstructionsArtifact",
    "ResourcesArtifact",
    "SolutionArtifact",
    "GradingArtifact",
    "Artifacts")


version = "1.4.1"


class Paths:
    DOT = Path()

    INSTRUCTIONS = Path("instructions")
    RESOURCES = Path("resources")
    SOLUTION = Path("solution")
    GRADING = Path("grading")

    ASSIGNMENT = Path("assignment")
    PROBLEM = Path("problem")
    ARTIFACTS = Path("artifacts")
    ASSETS = Path("assets")
    INCLUDE = Path("grade", "include")

    @classmethod
    def glob_assignments(cls, material_path: Path) -> Iterator[Path]:
        """Provide a unified search for assignments."""

        for path in material_path.joinpath(cls.ASSIGNMENT).glob("*/"):
            if path.is_dir():
                yield path


class Files:
    README = "README.md"
    ASSIGNMENT = "assignment.json"
    PROBLEM = "problem.json"
    GRADING = "grading.json"
    TESTS = "tests.py"
    INDEX = "index.json"


class Artifact:
    path: Path

    def __init__(self, path: Path):
        self.path = path


class InstructionsArtifact(Artifact):
    """Instructions and assets."""


class ResourcesArtifact(Artifact):
    """Assignment skeleton code."""


class SolutionArtifact(Artifact):
    """Example solution."""


class GradingArtifact(Artifact):
    """Contains metadata about the assignment and grading scripts."""

    @property
    def index_path(self) -> Path:
        return self.path.joinpath(Files.INDEX)


class Artifacts:
    """Bundled artifacts produced by curricula_compile."""

    instructions: InstructionsArtifact
    resources: ResourcesArtifact
    solution: SolutionArtifact
    grading: GradingArtifact

    def __init__(self, path: Path):
        """Initialize each artifact on its corresponding path"""

        self.instructions = InstructionsArtifact(path.joinpath(Paths.INSTRUCTIONS))
        self.resources = ResourcesArtifact(path.joinpath(Paths.RESOURCES))
        self.solution = SolutionArtifact(path.joinpath(Paths.SOLUTION))
        self.grading = GradingArtifact(path.joinpath(Paths.GRADING))



from typing import Any, Optional


class OrthogonalNone:
    """Custom none."""

    def __bool__(self):
        """Enable or default."""

        return False


none = OrthogonalNone()


def not_none(name: str, value: Any, default: Any = none) -> Any:
    """Return only if not none value."""

    if value is none:
        if default is not none:
            return default
        raise RuntimeError(f"can't find a valid source for {name}")
    return value


class Configurable:
    """Provide resolve on self."""

    @classmethod
    def getter_name(cls, field_name: str):
        return f"get_{field_name}"

    def __setattr__(self, key, value):
        """Don't allow overwriting with none."""

        if value is none:
            return
        super().__setattr__(key, value)

    def is_resolvable(
            self,
            field_name: str = None,
            field_getter_name: Optional[str] = none,
            local: Any = none) -> bool:
        """Determine whether one of the three is defined."""

        if local is not none:
            return True

        if field_name is not None and hasattr(self, field_name):
            return True

        if field_getter_name is none and field_name is not None:
            field_getter_name = Configurable.getter_name(field_name)
        if field_getter_name is not None and hasattr(self, field_getter_name):
            return True

        return False

    def resolve(
            self: Any,
            field_name: str = None,
            field_getter_name: Optional[str] = none,
            local: Any = none,
            default: Any = none) -> Any:
        """Check value, self.name, then self.get_name()."""

        if local is not none:
            return local

        # Check self
        if field_name is not None and hasattr(self, field_name):
            value = getattr(self, field_name)
            if value is not none:
                return value

        # Try getter
        if field_getter_name is none and field_name is not None:
            field_getter_name = Configurable.getter_name(field_name)

        if field_getter_name is not None and hasattr(self, field_getter_name):
            getter = getattr(self, field_getter_name)
            if callable(getter):
                value = getter()
                if value is not none:
                    return value

        # Return default if provided
        if default is not none:
            return default

        raise RuntimeError(f"can't find a valid source for {field_name or 'value'}")



import subprocess
import timeit
import time

from ..log import log
from .debug import get_source_location

from typing import Optional, Tuple, Callable, IO, TypeVar, Any
from dataclasses import dataclass, asdict, field
from contextlib import contextmanager
from functools import lru_cache
from pathlib import Path


@dataclass(eq=False)
class ProcessError:
    """Error that occurs during process runtime."""

    description: str
    error_number: Optional[int] = None

    @classmethod
    def from_os_error(cls, error: OSError) -> "ProcessError":
        """Create from OS error."""

        if error.errno == 8:
            description = "executable format error"
        else:
            description = "failed to run executable"

        return ProcessError(description=description, error_number=error.errno)

    def dump(self) -> dict:
        """Serialize."""

        return asdict(self)


T = TypeVar("T")


@lru_cache(maxsize=None)
def nullable(function: Callable[[Any], T]) -> Callable[[Optional[Any]], Optional[T]]:
    """None should pass through."""

    return lambda value: function(value) if value is not None else None


@dataclass(eq=False)
class ProcessCreation:
    """Information about how a process was started."""

    args: Tuple[str, ...]
    cwd: Optional[Path]

    def dump(self) -> dict:
        """Simple serialization."""

        dump = getattr(super(), "dump", dict)()
        dump.update(
            args=self.args,
            cwd=nullable(str)(self.cwd))
        return dump


@dataclass(eq=False)
class ProcessStreams:
    """Container for streamed data."""

    stdin: Optional[bytes] = None
    stdout: Optional[bytes] = None
    stderr: Optional[bytes] = None

    def dump(self) -> dict:
        """Decode any stream data from bytes."""

        dump = getattr(super(), "dump", dict)()
        dump.update(
            stdin=nullable(bytes.decode)(self.stdin),
            stdout=nullable(bytes.decode)(self.stdout),
            stderr=nullable(bytes.decode)(self.stderr))
        return dump


@dataclass(eq=False)
class Interaction(ProcessStreams, ProcessCreation):
    """Container for single interaction with a process."""

    elapsed: Optional[float] = None

    def dump(self) -> dict:
        """Make the runtime JSON serializable."""

        dump = super().dump()
        dump.update(elapsed=self.elapsed)
        return dump


@dataclass(eq=False)
class Runtime(ProcessStreams, ProcessCreation):
    """Runtime data extracted from running an external process."""

    # Elapsed time to finish
    elapsed: Optional[float] = None

    # A process must terminate one of three ways
    code: Optional[int] = None

    # Timeout to avoid hanging indefinitely
    timeout: Optional[float] = None
    timed_out: bool = False

    # Exception preventing start
    raised_exception: bool = False
    exception: Optional[ProcessError] = None

    def dump(self) -> dict:
        """Make the runtime JSON serializable."""

        dump = super().dump()
        dump.update(elapsed=self.elapsed)
        dump.update(code=self.code)
        dump.update(timeout=self.timeout)
        dump.update(timed_out=self.timed_out)
        dump.update(raised_exception=self.raised_exception)
        dump.update(exception=self.exception)
        return dump


@dataclass(eq=False)
class TimeoutExpired(RuntimeError):
    """Raised waiting for process termination or blocking I/O."""

    # Any extra data in the buffer
    buffer: bytes


@dataclass(eq=False)
class Stream:
    """Base class for a process stream wrapper."""

    file: IO[bytes]

    # Track all data passed through stream
    history: bytes = field(init=False, default=b"")


@dataclass(eq=False)
class Readable(Stream):
    """Custom IO stream for Interactive."""

    # Poll rate for reading
    POLL: float = 0.001

    def _read_block(self, condition: Callable[[bytes], bool] = None, timeout: float = None) -> Optional[bytes]:
        """Block until something besides None is returned."""

        buffer = b""

        timeout_time = None
        if timeout is not None:
            timeout_time = timeit.default_timer() + timeout

        while True:
            data = self.file.read()
            if data is not None:
                buffer += data
                if condition is None or condition(buffer):
                    break
            if timeout is not None and timeit.default_timer() >= timeout_time:
                self.history += buffer
                raise TimeoutExpired(buffer=buffer)
            time.sleep(self.POLL)

        self.history += buffer
        return buffer

    def read(
            self,
            condition: Callable[[bytes], bool] = None,
            timeout: float = None) -> Optional[bytes]:
        """Read from a stream.

        If blocking is disabled, may return None. If condition is not
        None, block until condition is satisfied. If timeout is not
        None, break after timeout and return buffer or None.
        """

        return self._read_block(condition=condition, timeout=timeout)


@dataclass(eq=False)
class Writable(Stream):
    """Provides a print-like wrapper for an output stream."""

    def write(
            self,
            *values: bytes,
            sep: bytes = b" ",
            end: bytes = b"\n",
            flush: bool = True):
        """Write to the stream like traditional print."""

        data = sep.join(values) + end
        self.file.write(data)
        self.history += data
        if flush:
            try:
                self.file.flush()
            except BrokenPipeError:
                pass


@dataclass(eq=False)
class Interactive:
    """An interactive runtime session."""

    _args: Tuple[str, ...]
    _process: subprocess.Popen
    _start_time: float
    cwd: Optional[Path]
    stdin: Writable
    stdout: Readable
    stderr: Readable

    _recording: Optional[Interaction] = None

    def __init__(self, args: Tuple[str, ...], cwd: Path = None):
        """Start up the new process."""

        self._args = args
        self._process = subprocess.Popen(
            args=args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            cwd=str(cwd) if cwd is not None else None)
        self.cwd = cwd
        self.stdin = Writable(self._process.stdin)
        self.stdout = Readable(self._process.stdout)
        self.stderr = Readable(self._process.stderr)
        self._start_time = timeit.default_timer()

    def poll(self) -> bool:
        """Check whether the interactive has terminated."""

        return self._process.poll() is None

    @contextmanager
    def recording(self) -> Interaction:
        """Record a frame of all process streams."""

        if self._recording is not None:
            raise RuntimeError("Cannot make multiple runtime recordings at once!")

        partial = Interaction(args=self._args, cwd=self.cwd)
        stdin_index = len(self.stdin.history)
        stdout_index = len(self.stdout.history)
        stderr_index = len(self.stderr.history)
        start_time = timeit.default_timer()

        yield partial

        # Collect everything that changed
        partial.elapsed = timeit.default_timer() - start_time
        partial.stdin = self.stdin.history[stdin_index:]
        partial.stdout = self.stdout.history[stdout_index:]
        partial.stderr = self.stderr.history[stderr_index:]

    def close(self, timeout: float = None) -> Runtime:
        """Block until exit."""

        raised_exception = False
        exception = None
        timed_out = False
        stdout = b""
        stderr = b""

        try:
            stdout, stderr = self._process.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            timed_out = True
        except OSError as error:
            raised_exception = True
            exception = ProcessError.from_os_error(error)

        stop_time = timeit.default_timer()
        return Runtime(
            args=self._args,
            cwd=self.cwd,
            timeout=timeout,
            code=self._process.returncode,
            elapsed=stop_time - self._start_time,
            stdin=self.stdin.history,
            stdout=self.stdout.history + stdout,
            stderr=self.stderr.history + stderr,
            raised_exception=raised_exception,
            exception=exception,
            timed_out=timed_out)


def run(*args: str, stdin: bytes = None, timeout: float = None, cwd: Path = None) -> Runtime:
    """Run an executable with a list of command line arguments.

    The provided path must be absolute in order to properly execute
    the program. Args provided are passed as they would be from the
    command line. The timeout is measured in seconds.

    The process_setup callable is invoked within the spawned process
    prior to the execution of the command.
    """

    if timeout is None:
        log.warning(f"process.run has been invoked without a timeout from {get_source_location()}")

    # Spawn the process, access stdout and stderr
    try:
        if stdin is not None:
            process = subprocess.Popen(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
                cwd=str(cwd) if cwd is not None else None)
        else:
            process = subprocess.Popen(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=str(cwd) if cwd is not None else None)

    # Catch common errors
    except OSError as error:
        exception = ProcessError.from_os_error(error)
        return Runtime(args=args, cwd=cwd, timeout=timeout, stdin=stdin, raised_exception=True, exception=exception)
    except ValueError:
        exception = ProcessError(description="failed to open process")
        return Runtime(args=args, cwd=cwd, timeout=timeout, stdin=stdin, raised_exception=True, exception=exception)
    except subprocess.SubprocessError as exception:
        exception = ProcessError(description=str(exception))
        return Runtime(args=args, cwd=cwd, timeout=timeout, stdin=stdin, raised_exception=True, exception=exception)

    # Wait for the process to finish with timeout
    start = timeit.default_timer()
    try:
        stdout, stderr = process.communicate(input=stdin, timeout=timeout)
    except subprocess.TimeoutExpired:
        process.kill()

        # Recover data
        try:
            stdout, stderr = process.communicate(timeout=1)
        except subprocess.TimeoutExpired:
            stdout, stderr = None, None

        return Runtime(args=args, cwd=cwd, timeout=timeout, stdin=stdin, stdout=stdout, stderr=stderr, timed_out=True)

    # Check elapsed
    elapsed = timeit.default_timer() - start
    return Runtime(
        args=args,
        cwd=cwd,
        timeout=timeout,
        code=process.returncode,
        elapsed=elapsed,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr)


def interact(*args: str) -> Interactive:
    """Shorthand for interactive, makes the interface nicer."""

    return Interactive(args=args)



import os
from xml.etree.ElementTree import Element, parse, ParseError
from typing import Optional, List
from dataclasses import dataclass, field
from pathlib import Path

from . import process

VALGRIND_ARGS = ("valgrind", "--tool=memcheck", "--leak-check=yes", "--xml=yes")
VALGRIND_XML_FILE = "valgrind.xml"


@dataclass
class ValgrindWhat:
    """Explanation of error, can have dynamic tags."""

    text: str
    fields: dict = field(default_factory=dict)

    @classmethod
    def load(cls, element: Optional[Element]) -> "Optional[ValgrindWhat]":
        """Load either what or xwhat."""

        if element is None:
            return None

        if element.tag == "what":
            return ValgrindWhat(element.text)
        else:
            text = ""
            fields = dict()
            for child in element:
                if child.tag == "tag":
                    text = child.text
                else:
                    fields[child.tag] = child.text
            return ValgrindWhat(text, fields)


@dataclass
class ValgrindError:
    """Represents an error tag from a Valgrind XML report."""

    unique: int
    tid: int
    kind: str
    what: Optional[ValgrindWhat]

    @classmethod
    def load(cls, element: Element) -> "ValgrindError":
        """Load an error from an element."""

        unique = int(element.find("unique").text, 16)
        tid = int(element.find("tid").text)
        kind = element.find("kind").text
        what = ValgrindWhat.load(element.find("what") or element.find("xwhat"))
        return cls(unique, tid, kind, what)


@dataclass
class ValgrindReport:
    """Include data about memory lost and errors."""

    runtime: process.Runtime
    valgrind_errors: Optional[List[ValgrindError]]
    error: str = None

    def memory_lost(self) -> (int, int):
        """Count up bytes and blocks lost."""

        leaked_blocks = 0
        leaked_bytes = 0
        for error in self.valgrind_errors:
            if error.kind in ("Leak_DefinitelyLost", "Leak_IndirectlyLost", "Leak_PossiblyLost"):
                leaked_blocks += int(error.what.fields["leakedblocks"])
                leaked_bytes += int(error.what.fields["leakedbytes"])
        return leaked_blocks, leaked_bytes


def run(*args: str, stdin: bytes = None, timeout: float = None, cwd: Path = None) -> Optional[ValgrindReport]:
    """Run valgrind on the program and return IR count."""

    runtime = process.run(
        *VALGRIND_ARGS,
        f"--xml-file={VALGRIND_XML_FILE}",
        *args,
        stdin=stdin,
        timeout=timeout,
        cwd=cwd)
    if os.path.exists(VALGRIND_XML_FILE):
        errors = []
        with open(VALGRIND_XML_FILE) as file:
            try:
                root = parse(file).getroot()
            except ParseError:
                return ValgrindReport(runtime, None, error="cannot parse valgrind xml")
            for child in root:
                if child.tag == "error":
                    errors.append(ValgrindError.load(child))
        os.remove(VALGRIND_XML_FILE)
        return ValgrindReport(runtime=runtime, valgrind_errors=errors)
    return ValgrindReport(runtime=runtime, valgrind_errors=None, error="valgrind did not write to output")



import jinja2
import logging
from decimal import Decimal
from pathlib import Path
from typing import Any, Dict

root = Path(__file__).absolute().parent
log = logging.getLogger("curricula")


def pretty(decimal: Decimal) -> str:
    """Display a number nicely."""

    if int(decimal) == decimal:
        return str(int(decimal))
    return str(round(decimal, 3)).rstrip("0")


def percentage(d: Any, digits: int = 1) -> str:
    """Convert a float to a nice-looking percentage."""

    converted = d if isinstance(d, Decimal) else Decimal(str(d))
    converted *= 100
    if converted == converted.to_integral_value():
        return f"{int(converted)}%"
    return f"{round(converted, digits)}%"


JINJA2_FILTERS = {
    "pretty": pretty,
    "datetime": lambda d: d.strftime("%B %d, %Y at %H:%M"),
    "date": lambda d: d.strftime("%B %d, %Y"),
    "percentage": percentage,
}


def jinja2_create_environment(
        default_template_path: Path,
        custom_template_path: Path = None,
        assignment_path: Path = None,
        problem_paths: Dict[str, Path] = None) -> jinja2.Environment:
    """Configure a jinja2 environment."""

    log.debug("creating jinja2 environment")

    # Create a loader in the order of arguments
    mapping = {}
    if assignment_path:
        mapping["assignment"] = jinja2.FileSystemLoader(str(assignment_path))
    if problem_paths:
        for key, path in problem_paths.items():
            mapping[key] = jinja2.FileSystemLoader(str(path))

    # Add custom templates
    mapping["template"] = jinja2.ChoiceLoader(
        ([jinja2.FileSystemLoader(str(custom_template_path))] if custom_template_path is not None else []) +
        [jinja2.FileSystemLoader(str(default_template_path))])
    loader = jinja2.PrefixLoader(mapping, delimiter=":")

    # Actually create the environment, we use square brackets to avoid
    # clashing with Jekyll
    environment = jinja2.Environment(
        loader=loader,
        block_start_string="[%",
        block_end_string="%]",
        variable_start_string="[[",
        variable_end_string="]]",
        comment_start_string="[#",
        comment_end_string="#]",
        autoescape=False,
        keep_trailing_newline=True)

    # Custom filters
    environment.filters.update(JINJA2_FILTERS)

    return environment



import importlib.util
from setuptools import setup, find_packages
from pathlib import Path


root = Path(__file__).absolute().parent

with root.joinpath("README.md").open() as fh:
    long_description = fh.read()

spec = importlib.util.spec_from_file_location("curricula", str(root.joinpath("curricula", "__init__.py")))
module = importlib.util.module_from_spec(spec)
spec.loader.exec_module(module)

setup(
    name="curricula",
    version=module.__version__,
    description="A content manager and grading toolkit for evaluating student code",
    url="https://github.com/curriculagg/curricula",
    author="Noah Kim",
    author_email="noahbkim@gmail.com",

    # Extra
    long_description=long_description,
    long_description_content_type="text/markdown",

    # Python
    python_requires=">=3.9",

    # Packaging
    packages=find_packages(),
    zip_safe=False)



import os
import shutil
import distutils.dir_util
from pathlib import Path


def contains(parent: Path, child: Path) -> bool:
    """Check if the child path is within the parent path."""

    for a, b in zip(parent.resolve().parts, child.resolve().parts):
        if a is None:
            return True
        if a != b:
            return False
    return True


def relative(root: Path, path: Path):
    """Left-truncate the path by the root."""

    return Path(*path.parts[len(root.parts):])


def move(source: Path, destination: Path):
    """Rename a file."""

    shutil.move(str(source), str(destination))


def copy(source: Path, destination: Path):
    """Copy a file."""

    if source.is_file():
        copy_file(source, destination)
    else:
        copy_directory(source, destination)


def copy_file(source: Path, destination: Path):
    """Copy a file."""

    shutil.copy(str(source), str(destination))


def copy_directory(source: Path, destination: Path, merge: bool = False):
    """Copy all files recursively."""

    if merge:
        distutils.dir_util.copy_tree(str(source), str(destination))
    else:
        if destination.exists():
            delete(destination)
        shutil.copytree(str(source), str(destination))


def delete(path: Path):
    """Delete a file or directory."""

    if path.is_file():
        delete_file(path)
    else:
        delete_directory(path)


def delete_file(path: Path):
    """Delete a file."""

    os.remove(str(path))


def delete_directory(path: Path):
    """Delete a directory recursively."""

    shutil.rmtree(str(path))


def replace_directory(path: Path):
    """Make sure a directory is present and empty."""

    if path.is_file() or path.is_dir():
        delete(path)
    path.mkdir(parents=True)


def add_mode(path: Path, mode: int):
    """Do chmod and add a mode."""

    os.chmod(str(path), os.stat(str(path)).st_mode | mode)


def subtract_mode(path: Path, mode: int):
    """Do chmod and subtract a mode."""

    os.chmod(str(path), os.stat(str(path)).st_mode & ~mode)



import os
import tempfile
from pathlib import Path
from typing import Optional, Tuple

from . import process
from .files import delete_file

__all__ = ("count",)


def read_last_line(path: Path) -> Optional[str]:
    """IR count appears at the end of the callgrind output."""

    with path.open("rb") as file:
        if len(file.read(2)) < 2:
            return None

        file.seek(0)
        file.seek(-2, os.SEEK_END)
        while file.read(1) != b"\n":
            file.seek(-2, os.SEEK_CUR)
            if file.tell() == 0:
                return None
        return file.readlines()[-1].decode()


def count(
        *args: str,
        stdin: bytes = None,
        timeout: float = None,
        cwd: Path = None,
        function_name: str = None) -> Tuple[process.Runtime, Optional[int]]:
    """Run callgrind on the program and return IR count."""

    extra_valgrind_args = []
    if function_name is not None:
        extra_valgrind_args.append(f"--toggle-collect={function_name}")

    out_path = Path(next(tempfile._get_candidate_names()))
    runtime = process.run(
        "valgrind",
        "--tool=callgrind",
        f"--callgrind-out-file={out_path.parts[-1]}",
        *extra_valgrind_args,
        *args,
        stdin=stdin,
        timeout=timeout,
        cwd=cwd)
    if out_path.exists():
        last_line = read_last_line(out_path)
        if last_line is None:
            return runtime, None
        result = int(last_line.rsplit(maxsplit=1)[1])
        delete_file(out_path)
        return runtime, result
    return runtime, None
